@@ -6,6 +6,7 @@ import numpy as np
 import scipy.stats
 import scipy.special as sc
 import scipy.optimize as so
+import decimal
 from decimal import *
 import math
 import functools
@@ -120,12 +121,22 @@ class StatisticalHelperMethods:
     
     @staticmethod
     def precise_beta(a: float, b: float) -> Decimal:
-        return (StatisticalHelperMethods.precise_gamma(a)*StatisticalHelperMethods.precise_gamma(b))/(StatisticalHelperMethods.precise_gamma(a+b))
+        precise_beta_function = lambda a,b: (StatisticalHelperMethods.precise_gamma(a)*StatisticalHelperMethods.precise_gamma(b))/(StatisticalHelperMethods.precise_gamma(a+b))
+        try:
+            return precise_beta_function(a,b)
+        except decimal.Overflow:
+            print("Decimal overflow thrown - attempting calculatin with doubled the precision from {} to {}".format(decimal.getcontext().prec, decimal.getcontext().prec*2))
+            with decimal.localcontext() as ctx:
+                ctx.prec *= 2  # increase the precision by 10
+                return precise_beta_function(a,b)
+
+
 
     # TODO: Change KL to be base e to be between 0 and 1.
     @staticmethod
     def KL_beta(alpha_f: float, beta_f: float, alpha_g: float, beta_g: float):
         # NOTE: Because the beta function can output extremely small values, using Decimal for higher precision
+
         return float(Decimal.ln(StatisticalHelperMethods.precise_beta(alpha_g, beta_g)/(StatisticalHelperMethods.precise_beta(alpha_f, beta_f)))) + ((alpha_f - alpha_g)*(sc.digamma(alpha_f) - sc.digamma(alpha_f+beta_f))) + ((beta_f - beta_g)*(sc.digamma(beta_f) - sc.digamma(alpha_f+beta_f)))
 
     @staticmethod
@@ -198,6 +209,17 @@ def perform_singleton_score_imputation(each_guide: Guide,
     
     imputation_posterior_beta = (singleton_imputation_prior_strength*negative_control_guide_pop2_total_normalized_counts_reps[replicate_indices])
 
+    max_imputation_posterior = 5000 # Setting this max, since calculating the KL divergence of a very high posterior value will result in a precision error
+    for rep_i in range(len(replicate_indices)):
+        if max(imputation_posterior_alpha[rep_i],imputation_posterior_beta[rep_i]) > max_imputation_posterior:
+            print("Downscaling posterior: {}, {}".format(imputation_posterior_alpha[rep_i],imputation_posterior_beta[rep_i]))
+            if imputation_posterior_alpha[rep_i] > imputation_posterior_beta[rep_i]:
+                downscale_factor = max_imputation_posterior/imputation_posterior_alpha[rep_i]
+                imputation_posterior_alpha[rep_i] = imputation_posterior_alpha[rep_i] * downscale_factor
+                imputation_posterior_beta[rep_i] = imputation_posterior_beta[rep_i] * downscale_factor
+
+
+
     return imputation_posterior_alpha, imputation_posterior_beta
 
 def perform_neighboorhood_score_imputation(each_guide: Guide, 
@@ -259,6 +281,7 @@ def optimize_singleton_imputation_prior_strength(
             true_beta = each_guide.pop2_normalized_count_reps[rep_i]
 
             # Calculate KL divergence between the posterior and the likelihood
+            
             KL_guide_imputation_score: float = StatisticalHelperMethods.KL_beta(true_alpha, true_beta, imputation_posterior_alpha, imputation_posterior_beta)
 
             # Add score to the main placeholder to get the final sum
@@ -268,15 +291,16 @@ def optimize_singleton_imputation_prior_strength(
 
 
     def optimize_singleton_imputation_model_weights(rep_i, params):
-        KL_guide_imputation_score_total_negative, negative_n = retrieve_objective_of_guide_set(rep_i, experiment_guide_sets.negative_control_guides, params)
-        KL_guide_imputation_score_total_positive, positive_n = retrieve_objective_of_guide_set(rep_i, experiment_guide_sets.positive_control_guides, params)
-        KL_guide_imputation_score_total_observation, observation_n = retrieve_objective_of_guide_set(rep_i, experiment_guide_sets.observation_guides, params)
-
-        KL_guide_imputation_score_total_combined_avg = (KL_guide_imputation_score_total_negative+KL_guide_imputation_score_total_positive+KL_guide_imputation_score_total_observation)/(negative_n+positive_n+observation_n)
-            
-        KL_guide_imputation_score_total_negative_avg = np.inf if negative_n == 0 else KL_guide_imputation_score_total_negative/negative_n
-        KL_guide_imputation_score_total_positive_avg = np.inf if positive_n == 0 else KL_guide_imputation_score_total_positive/positive_n
-        KL_guide_imputation_score_total_observation_avg = np.inf if observation_n == 0 else KL_guide_imputation_score_total_observation/observation_n
+        KL_guide_imputation_score_total_negative, negative_total_n = retrieve_objective_of_guide_set(rep_i, experiment_guide_sets.negative_control_guides, params)
+        KL_guide_imputation_score_total_positive,  positive_total_n = retrieve_objective_of_guide_set(rep_i, experiment_guide_sets.positive_control_guides, params)
+        KL_guide_imputation_score_total_observation,  observation_total_n = retrieve_objective_of_guide_set(rep_i, experiment_guide_sets.observation_guides, params)
+        
+        total_n = (negative_total_n+positive_total_n+observation_total_n)
+        KL_guide_imputation_score_total_combined_avg = np.inf if total_n == 0 else (KL_guide_imputation_score_total_negative + KL_guide_imputation_score_total_positive + KL_guide_imputation_score_total_observation) / total_n
+        
+        KL_guide_imputation_score_total_negative_avg = np.inf if negative_total_n == 0 else KL_guide_imputation_score_total_negative/negative_total_n
+        KL_guide_imputation_score_total_positive_avg = np.inf if positive_total_n == 0 else KL_guide_imputation_score_total_positive/positive_total_n
+        KL_guide_imputation_score_total_observation_avg = np.inf if observation_total_n == 0 else KL_guide_imputation_score_total_observation/observation_total_n
 
 
         return KL_guide_imputation_score_total_combined_avg, KL_guide_imputation_score_total_negative_avg, KL_guide_imputation_score_total_positive_avg, KL_guide_imputation_score_total_observation_avg
@@ -351,7 +375,7 @@ def optimize_neighborhood_imputation_prior_strength(
     def retrieve_objective_of_guide_set(rep_i, guide_set: List[Guide], params):
         spatial_imputation_prior_strength_test, spatial_imputation_likelihood_strength_test = params
         if len(guide_set) == 0:
-            return 0, 0
+            return 0,0,0
 
         KL_guide_imputation_score_total: float = 0
 
@@ -383,23 +407,28 @@ def optimize_neighborhood_imputation_prior_strength(
             # Add score to the main placeholder to get the final sum
             KL_guide_imputation_score_total = KL_guide_imputation_score_total + KL_guide_imputation_score 
             
-        return KL_guide_imputation_score_total, len(guide_set)
+        return KL_guide_imputation_score_total, neighborhood_optimization_guide_sample_size_i, len(guide_set)
 
     def optimize_neighboorhood_imputation_model_weights(rep_i, params):
         
          # Iterate through each guide to test prior with tested weight
 
-        KL_guide_imputation_score_total_negative, negative_n = retrieve_objective_of_guide_set(rep_i, experiment_guide_sets.negative_control_guides, params)
+        KL_guide_imputation_score_total_negative, negative_sampled_n, negative_total_n = retrieve_objective_of_guide_set(rep_i, experiment_guide_sets.negative_control_guides, params)
 
-        KL_guide_imputation_score_total_positive, positive_n = retrieve_objective_of_guide_set(rep_i, experiment_guide_sets.positive_control_guides, params)
+        KL_guide_imputation_score_total_positive, positive_sampled_n, positive_total_n = retrieve_objective_of_guide_set(rep_i, experiment_guide_sets.positive_control_guides, params)
         
-        KL_guide_imputation_score_total_observation, observation_n = retrieve_objective_of_guide_set(rep_i, experiment_guide_sets.observation_guides, params)
+        KL_guide_imputation_score_total_observation, observation_sampled_n, observation_total_n = retrieve_objective_of_guide_set(rep_i, experiment_guide_sets.observation_guides, params)
         
-        KL_guide_imputation_score_total_combined_avg = (KL_guide_imputation_score_total_negative+KL_guide_imputation_score_total_positive+KL_guide_imputation_score_total_observation)/(negative_n+positive_n+observation_n)
+        total_n = (negative_total_n+positive_total_n+observation_total_n)
+        KL_guide_imputation_score_total_combined_avg = np.inf if total_n == 0 else (
+            (0 if negative_sampled_n == 0 else (KL_guide_imputation_score_total_negative/negative_sampled_n)*negative_total_n) + 
+            (0 if positive_sampled_n == 0 else (KL_guide_imputation_score_total_positive/positive_sampled_n)*positive_total_n) + 
+            (0 if observation_sampled_n == 0 else (KL_guide_imputation_score_total_observation/observation_sampled_n)*observation_total_n)
+            ) / total_n
         
-        KL_guide_imputation_score_total_negative_avg = np.inf if negative_n == 0 else KL_guide_imputation_score_total_negative/negative_n
-        KL_guide_imputation_score_total_positive_avg = np.inf if positive_n == 0 else KL_guide_imputation_score_total_positive/positive_n
-        KL_guide_imputation_score_total_observation_avg = np.inf if observation_n == 0 else KL_guide_imputation_score_total_observation/observation_n
+        KL_guide_imputation_score_total_negative_avg = np.inf if negative_sampled_n == 0 else KL_guide_imputation_score_total_negative/negative_sampled_n
+        KL_guide_imputation_score_total_positive_avg = np.inf if positive_sampled_n == 0 else KL_guide_imputation_score_total_positive/positive_sampled_n
+        KL_guide_imputation_score_total_observation_avg = np.inf if observation_sampled_n == 0 else KL_guide_imputation_score_total_observation/observation_sampled_n
 
         return KL_guide_imputation_score_total_combined_avg, KL_guide_imputation_score_total_negative_avg, KL_guide_imputation_score_total_positive_avg, KL_guide_imputation_score_total_observation_avg
 
